/////////////////////////////////////////
// Scripts by Tom N Harris (Telliamed)
// mailto:telliamed@whoopdedo.cjb.net
//
// Interaction scripts involve UI enhancements or complex AI behaviors.
//

#ifndef INTERACTSCRIPT_H
#define INTERACTSCRIPT_H

#if _MSC_VER > 1000
#pragma once
#endif

#include "lg/scrservices.h"

#include "scriptvars.h"
#include "Script.h"
#include "GenericScript.h"
#include "GenericTrap.h"
#include "GenericTrigger.h"


/*************************
 * KnockOnDoor
 *
 * Signals AI with ``knock_knock`` when a closed & locked door is frobbed.
 * Will also send an Interact message with the data ''knock_knock'' as an
 * alternate way to respond. (Since AI can only have one signal response.)
 *
 * Messages: FrobWorldEnd
 * Links: Owns(AIToSignal, DoorObject)
 * Signals: knock_knock
 */
DECLARE_SCRIPT(cScr_KnockOnDoor,cScript)
{
public:
	DECLARE_FACTORY(KnockOnDoor,CustomScript);
	cScr_KnockOnDoor(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnFrob);
	static int LinkIterFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData);

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("FrobWorldEnd",cScr_KnockOnDoor::OnFrob);
	}
};


class cDisplayBookText
{
public:
	cDisplayBookText(int iHostObjId)
		: m_iObjId(iHostObjId)
	{ }

private:
	int m_iObjId;

protected:
	virtual int DisplayBookPage(const char* pszBook, int iPage);
	virtual int DisplayText(const char* pszText, int iColor, int iTime);
	virtual int DefaultTextColor(void);
	virtual int ChangeBookPage(const char* pszBook, int iPage, int iTime);
};

/**************************
 * OnScreenText
 * 
 * Display text from a book file on-screen. Supports multiple pages, color, 
 * and auto-scroll. The page to display is specified in the parameter page. 
 * The parameter is modified by the script. The text for a page is in the 
 * book string ``page_0``. There is no effective limit to the number of pages.
 * The color of a page is ``page_0_color`` as #AABBCC (HTML format). The time 
 * to display a page is ``page_0_time``. Time is specified as milliseconds, 
 * or as seconds by appending 's', or as minutes by appending 'm'. So 
 * $$1m$$ = $$60s$$ = $$60000$$. ``page_0_next`` specifies which page to display next.
 * If not specified, the next page is the next in-order, or 0 if it doesn't 
 * exist. ``page_0_auto`` is the time before the next page is automatically 
 * displayed. The auto-scroll time runs concurrently with the display time,
 * so $$page_0_time: 3s$$ and $$page_0_auto: 4s$$ with display the text for 3 seconds,
 * then 1 second later will show the next page.
 *
 * ``ScriptControl`` with just a number will set the current page but not display 
 * it. You can also use a string for greater control. Use //<op><num>//.
 * ^ Op. ^  Result  ^
 * |  @  | Set the page without displaying it.  |
 * |  #  | Set the page then display it.  |
 * |  +  | Increment the page counter.  |
 * |  -  | Decrement the page counter.  |
 *   
 * ``TurnOff`` does nothing. ``ScriptControl`` will set the current page but doesn't
 * display it.
 * The default time is 5 seconds. The default color is white.
 *
 * Inherits: GenericScript
 * Parameters: page
 * Properties: Book
 * Strings: page_n, page_n_color, page_n_time, page_n_auto, page_n_next
 */
DECLARE_SCRIPT(cScr_OnScreenText,cScrGeneric), cDisplayBookText
{
public:
	DECLARE_FACTORY(OnScreenText,GenericScript);
	cScr_OnScreenText(const char* pszName, int iHostObjId)
		: cScrGeneric(pszName, iHostObjId),
		  cDisplayBookText(iHostObjId),
		  SCRIPT_VAROBJ(cScr_OnScreenText,m_hAutoScroll,iHostObjId)
	{ }

private:
	script_handle<tScrTimer> m_hAutoScroll;

	DECLARE_MSGHANDLER(OnBeginScript)
	{
		m_hAutoScroll.Init();
		return 1;
	}
	DECLARE_MSGHANDLER(OnTimer);

	int DisplayPage(int iPage);
	int ChangePage(int iPage);

protected:
	virtual int ChangeBookPage(const char* pszBook, int iPage, int iTime);

protected:
	virtual long OnTurnOn(sScrMsg* pMsg, sMultiParm* pReply);
	virtual long OnTurnOff(sScrMsg* pMsg, sMultiParm* pReply);
	virtual long OnControl(sScrMsg* pMsg, sMultiParm* pReply);
	virtual void InitScriptData()
	{
		cScrGeneric::InitScriptData();
		REGISTER_MSGHANDLER("BeginScript",cScr_OnScreenText::OnBeginScript);
		REGISTER_MSGHANDLER("Timer",cScr_OnScreenText::OnTimer);
	}
};


/*************************
 * InteractFrobHack
 *
 * Replacement for ``CorpseFrobHack`` that allows you to frob an active AI.
 * You need the meta-property ``FrobInteract`` with the appropriate frob
 * settings. Usually set to $$World: Script$$.
 * This script duplicates all the normal behavior of ``CorpseFrobHack``,
 * including the "Loot and Lug" (but without the annoying mono spew).
 *
 * Messages: AIModeChange, Slain, FrobWorldEnd
 * MetaProperties: FrobInteract
 */
DECLARE_SCRIPT(cScr_InteractFrobHack,cScript)
{
public:
	DECLARE_FACTORY(InteractFrobHack,CustomScript);
	cScr_InteractFrobHack(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnBeginScript);
	DECLARE_MSGHANDLER(OnAIModeChange);
	DECLARE_MSGHANDLER(OnSlain);
	DECLARE_MSGHANDLER(OnFrobWorldEnd);

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("BeginScript",cScr_InteractFrobHack::OnBeginScript);
		REGISTER_MSGHANDLER("AIModeChange",cScr_InteractFrobHack::OnAIModeChange);
		REGISTER_MSGHANDLER("Slain",cScr_InteractFrobHack::OnSlain);
		REGISTER_MSGHANDLER("FrobWorldEnd",cScr_InteractFrobHack::OnFrobWorldEnd);
	}
};


/*************************
 * AIInteraction
 *
 * Starts a conversation in response to an ``Interact`` message.
 * The object to trigger is chosen from any number of ``ScriptParams``
 * links that have the same data as was sent with the ``Interact`` message.
 * If no links with the specific data are found, it will pick a 
 * generic link with the data "$$Interact$$".
 *
 * Messages: Interact, SetInteraction
 * Links: ScriptParams
 */
DECLARE_SCRIPT(cScr_AIInteraction,cScript)
{
public:
	DECLARE_FACTORY(AIInteraction,CustomScript);
	cScr_AIInteraction(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnInteract);
	DECLARE_MSGHANDLER(OnSetInteraction);

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("Interact",cScr_AIInteraction::OnInteract);
		REGISTER_MSGHANDLER("SetInteraction",cScr_AIInteraction::OnSetInteraction);
	}
};


/*************************
 * PlayerInteraction
 *
 * When the object is tool-frobbed, an Interact message is sent 
 * to the destination of the frob. The ``GameName`` of the object
 * is sent with the message.
 * The ``new_speech`` schema will be played when the object
 * is contained by an avatar.
 *
 * Messages: FrobToolEnd, Contained
 */
DECLARE_SCRIPT(cScr_PlayerInteraction,cScript)
{
public:
	DECLARE_FACTORY(PlayerInteraction,CustomScript);
	cScr_PlayerInteraction(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnToolFrob);
	DECLARE_MSGHANDLER(OnContained);

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("FrobToolEnd",cScr_PlayerInteraction::OnToolFrob);
		REGISTER_MSGHANDLER("Contained",cScr_PlayerInteraction::OnContained);
	}
};


/*************************
 * AwarenessFilter
 *
 * When placed on another object, an AI can send a ``CanISeeYou`` or 
 * ``CanIHearYou`` message to test for awareness of the object from the
 * AI. When placed on an AI, it can send ``CanISeePlayer`` or ``CanIHearPlayer``
 * messages to itself to test for awareness of the ``Player`` object. Or it 
 * can send ``CanISeeObject`` or ``CanIHearObject`` with the object specified
 * in the first data parameter.
 *
 * Messages: CanISeeYou, CanIHearYou, CanISeeObject, CanIHearObject,
 *           CanISeePlayer, CanIHearPlayer
 */
DECLARE_SCRIPT(cScr_AwarenessFilter,cScript)
{
public:
	DECLARE_FACTORY(AwarenessFilter,CustomScript);
	cScr_AwarenessFilter(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnCanISeeYou);
	DECLARE_MSGHANDLER(OnCanIHearYou);
	DECLARE_MSGHANDLER(OnCanISeeObject);
	DECLARE_MSGHANDLER(OnCanIHearObject);
	DECLARE_MSGHANDLER(OnCanISeePlayer);
	DECLARE_MSGHANDLER(OnCanIHearPlayer);

	bool TestAwareness(int from, int to, int flags);

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("CanISeeYou",cScr_AwarenessFilter::OnCanISeeYou);
		REGISTER_MSGHANDLER("CanIHearYou",cScr_AwarenessFilter::OnCanIHearYou);
		REGISTER_MSGHANDLER("CanISeeObject",cScr_AwarenessFilter::OnCanISeeObject);
		REGISTER_MSGHANDLER("CanIHearObject",cScr_AwarenessFilter::OnCanIHearObject);
		REGISTER_MSGHANDLER("CanISeePlayer",cScr_AwarenessFilter::OnCanISeePlayer);
		REGISTER_MSGHANDLER("CanIHearPlayer",cScr_AwarenessFilter::OnCanIHearPlayer);
	}
};


/*************************
 * TrapSignalAwareness
 *
 * Signal all AI with an awareness of the ``ControlDevice`` linked objects.
 *
 * Inherits: GenericTrap
 * Parameters: signal, awareness(0 = any, 1 = sight only, 2 = sound only, 3 = both)
 * Links: ControlDevice
 */
DECLARE_SCRIPT(cScr_SignalAwareness,cScrTrap)
{
public:
	DECLARE_FACTORY(TrapSignalAwareness,GenericTrap);
	cScr_SignalAwareness(const char* pszName, int iHostObjId)
		: cScrTrap(pszName,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnSim);

	static int AwareObjIter(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData);

protected:
	virtual long OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply);
	virtual void InitScriptData()
	{
		cScrTrap::InitScriptData();
		REGISTER_MSGHANDLER("Sim",cScr_SignalAwareness::OnSim);
	}
};


/*************************
 * PrivateRoom
 *
 * Allows an AI that is normally non-hostile to become alarmed if it
 * sees the ``Player`` in a room. When the ``Player`` enters the room, a 
 * meta-property (``M-PlayerBreach``) is added to the ``Player``, and ``AIWatchObj``
 * links are created from every ``Human`` object to the ``Player``. If any AI
 * can see the player at the moment he enters the room, then the AI
 * is signalled with ``alarm``. When the ``Player`` exits the room, then 
 * the meta-property is removed and the ``AIWatchObj`` links are destroyed.
 * There is a slight pause (``exit_time`` param) after leaving a room before 
 * the links are destroyed. The idea is that an AI will be alerted if it 
 * sees that you've just walked out of the room, even if he didn't actually 
 * see you in the room itself. Another awareness test will be made when 
 * the links are removed.
 *
 * The ``M-PlayerBreach`` meta-property should have AI watch link defaults
 * set to $$PlayerEntry$$ with $$line-of-sight$$. It should also have a small 
 * but non-zero radius, and an awareness requirement of $$2$$. The 
 * [[AwarenessFilter]] script may also be useful.
 *
 * Messages: PlayerRoomEnter, PlayerRoomExit, Timer
 * Links: AIAwareness, AIWatchObj
 * MetaProperties: M-PlayerBreach
 * Signals: alarm
 */
DECLARE_SCRIPT(cScr_PrivateRoom,cScript)
{
public:
	DECLARE_FACTORY(PrivateRoom,CustomScript);
	cScr_PrivateRoom(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId),
		SCRIPT_VAROBJ(cScr_PrivateRoom,m_bPlayerBreach,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnBeginScript)
	{
		m_bPlayerBreach.Init();
		return 0;
	}
	DECLARE_MSGHANDLER(OnRoomEnter);
	DECLARE_MSGHANDLER(OnRoomExit);
	DECLARE_MSGHANDLER(OnTimer);

	script_int m_bPlayerBreach;

	void TestAwareness(int iMoveObjId);

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("PlayerRoomEnter",cScr_PrivateRoom::OnRoomEnter);
		REGISTER_MSGHANDLER("PlayerRoomExit",cScr_PrivateRoom::OnRoomExit);
		REGISTER_MSGHANDLER("Timer",cScr_PrivateRoom::OnTimer);
	}
};


/*************************
 * VendorRelay
 *
 * Will relay ``TurnOn``
 *
 * Loot cost can be given as the total value, or as three numbers for gold, gems, and goods.
 *
 * Inherits: GenericTrigger
 * Links: ControlDevice, ScriptParams("ErrorOutput")
 * Parameters: vendor("gold_only", "gems_only", "goods_only", "gold_first", "even", "exact"),
 *             loot(cost)
 */
DECLARE_SCRIPT(cScr_VendorRelay,cScript), cScrTrigger
{
public:
	DECLARE_FACTORY(VendorRelay,CustomScript);
	cScr_VendorRelay(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId), cScrTrigger(iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnRelay);

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("TurnOn",cScr_VendorRelay::OnRelay);
	}
};


/*************************
 * LootExchange
 *
 * On ``TurnOn``, converts an amount of the Player's loot from gems/goods into gold.
 *
 * The exchange rate is a real number and can be given as a fraction.
 *
 * Inherits: GenericTrap
 * Links: ControlDevice, ScriptParams("ErrorOutput")
 * Parameters: vendor("gems_only", "goods_only", "even"), exchange(fraction)
 */
DECLARE_SCRIPT(cScr_LootExchange,cScrTrap)
{
public:
	DECLARE_FACTORY(LootExchange,GenericTrap);
	cScr_LootExchange(const char* pszName, int iHostObjId)
		: cScrTrap(pszName,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnSim);

protected:
	virtual long OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply);
	virtual void InitScriptData()
	{
		cScrTrap::InitScriptData();
		REGISTER_MSGHANDLER("Sim",cScr_LootExchange::OnSim);
	}
};


/*************************
 * PurchaseItem
 *
 * When turned on, the Item object is duplicated and added to the $$Cart$$ 
 * container. If ``Trap\Quest Var`` is set, then the purchase price of the item 
 * is added to the variable.
 * On ``TurnOff``, the $$Item$$ is subtracted (by ``StackCount``) from the $$Cart$$
 * and the purchase price deducted from the qvar.
 * On ``Reset``, it removes all of its items from the $$Cart$$.
 * If the $$Item$$ is concrete, then the ``StackCount`` is the total inventory 
 * that is available, and each purchase is always one item.
 * You should link to the object with the [[ShoppingCart]] script using
 * a ``ScriptParams`` link of type $$Store$$. If that link exists, the ``Traq\Quest Var``
 * property will be read from the store object.
 *
 * Inherits: GenericTrap
 * Links: ScriptParams("Cart"), ScriptParams("Item"), ScriptParams("Store")
 * Properties: Trap\Quest Var
 * Messages: TurnOn, TurnOff, Reset
 */
DECLARE_SCRIPT(cScr_PurchaseItem,cScrTrap)
{
public:
	DECLARE_FACTORY(PurchaseItem,GenericTrap);
	cScr_PurchaseItem(const char* pszName, int iHostObjId)
		: cScrTrap(pszName,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnSim);
	DECLARE_MSGHANDLER(OnReset);

protected:
	virtual long OnControl(sScrMsg* pMsg, sMultiParm* pReply);
	virtual long OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply);
	virtual void InitScriptData()
	{
		cScrTrap::InitScriptData();
		REGISTER_MSGHANDLER("Sim",cScr_PurchaseItem::OnSim);
		REGISTER_MSGHANDLER("Reset",cScr_PurchaseItem::OnReset);
	}
};


/*************************
 * ShoppingCart
 *
 * Ignores the cart if the link doesn't exist. (Useful when using
 * the script as a lock.)
 * When turned off, the script will send ``Reset`` to all objects that 
 * link to this one with $$Store$$. It will also clear the contents of the cart
 * or, if no cart is specified, it sends ``Reset`` to all object that 
 * link to this one with $$Cart$$.
 *
 * Inherits: GenericTrap
 * Links: ScriptParams("Cart"), ScriptParams("Store")
 * Properties: Trap\Quest Var, Engine Features\Locked
 * Parameters: vendor
 */
DECLARE_SCRIPT(cScr_ShoppingCart,cScrTrap)
{
public:
	DECLARE_FACTORY(ShoppingCart,GenericTrap);
	cScr_ShoppingCart(const char* pszName, int iHostObjId)
		: cScrTrap(pszName,iHostObjId),
		SCRIPT_VAROBJ(cScr_ShoppingCart,m_iReallyLocked,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnBeginScript);
	DECLARE_MSGHANDLER(OnEndScript);
	DECLARE_MSGHANDLER(OnSim);
	DECLARE_MSGHANDLER(OnQuestChange);

	script_int m_iReallyLocked;

protected:
	virtual long OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply);
	virtual bool IsLocked(sScrMsg* pMsg)
		{ return m_iReallyLocked; }
	virtual void SetLock(bool bLock)
		{ m_iReallyLocked = static_cast<int>(bLock); }
	virtual void InitScriptData()
	{
		cScrTrap::InitScriptData();
		REGISTER_MSGHANDLER("Sim",cScr_ShoppingCart::OnSim);
		REGISTER_MSGHANDLER("BeginScript",cScr_ShoppingCart::OnBeginScript);
		REGISTER_MSGHANDLER("EndScript",cScr_ShoppingCart::OnEndScript);
		REGISTER_MSGHANDLER("QuestChange",cScr_ShoppingCart::OnQuestChange);
	}
};


/*******************
 * LootStats
 *
 * Tracks loot that is acquired/spent during a mission.
 *
 * Messages: Container
 */
DECLARE_SCRIPT(cScr_LootTracker,cScript)
{
public:
	DECLARE_FACTORY(LootStats,CustomScript);
	cScr_LootTracker(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId),
		SCRIPT_VAROBJ(cScr_LootTracker,m_bTracking,iHostObjId),
		SCRIPT_VAROBJ(cScr_LootTracker,m_iCurrentLoot,iHostObjId)
	{ }

private:
	script_int m_bTracking;
	script_int m_iCurrentLoot;

	DECLARE_MSGHANDLER(OnBeginScript);
	DECLARE_MSGHANDLER(OnSim);
	DECLARE_MSGHANDLER(OnTimer);
	DECLARE_MSGHANDLER(OnContainer);
	
protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("BeginScript", cScr_LootTracker::OnBeginScript);
		REGISTER_MSGHANDLER("Sim", cScr_LootTracker::OnSim);
		REGISTER_MSGHANDLER("Timer", cScr_LootTracker::OnTimer);
		REGISTER_MSGHANDLER("Container", cScr_LootTracker::OnContainer);
	}
};


/*******************
 * AIFactory
 *
 * Creates a random AI.
 * Properties and links from this object will be copied to the AI.
 *
 * Messages: Sim
 * Links: ScriptParams("Spawn")
 */
DECLARE_SCRIPT(cScr_Spawn, cScript)
{
public:
	DECLARE_FACTORY(AIFactory,CustomScript);
	cScr_Spawn(const char* pszName, int iHostObjId)
		: cScript(pszName, iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnSim);

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("Sim",cScr_Spawn::OnSim);
	}
};


#endif // INTERACTSCRIPT_H
