/////////////////////////////////////////
// Scripts by Tom N Harris (Telliamed)
// mailto:telliamed@whoopdedo.cjb.net
//
// NewScripts are implemented as hooks on one of the original 
// game scripts. Well, some of them are, anyway. 
//

#ifndef NEWSCRIPT_H
#define NEWSCRIPT_H

#if _MSC_VER > 1000
#pragma once
#endif

#include "lg/scrservices.h"

#include "scriptvars.h"
#include "Script.h"


/********************
 * NewCrystal
 * 
 * Substitute this object for another when contained by the player.
 * Acts like the standard ``Crystal`` script, but lets you specify what the
 * arrow (or whatever) will be.
 * The link can be on an archetype and the script will find it, so you
 * can make custom arrows as easily as regular ones. Cool, huh?
 *
 * Messages: Contained
 * Links: ScriptParams(Crystal, ArrowArchetype, "Crystal")
 */
DECLARE_SCRIPT(cScr_NewCrystal,cScript)
{
public:
	DECLARE_FACTORY(NewCrystal,CustomScript);
	cScr_NewCrystal(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnContained);
	long TransmogrifyCrystal(IObjectSrv* pObjSrv, int iDest);

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("Contained",cScr_NewCrystal::OnContained);
	}
};


/********************
 * NewKey
 *
 * Patch for ``StdKey`` to fix the ``PlayerToolFrob`` bug.
 *
 */
DECLARE_SCRIPT(cScr_NewKey,cScript)
{
public:
	DECLARE_FACTORY(NewKey,HookScript);
	cScr_NewKey(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId), m_pStdKey(NULL)
	{ }
	virtual ~cScr_NewKey ();

private:
	DECLARE_MSGHANDLER(OnFrobToolEnd);
	IScript* m_pStdKey;

protected:
	virtual void InitScriptData();
	virtual long __stdcall ReceiveMessage(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace);
};


/********************
 * NewGasLight
 *
 * Patch for ``GasLight`` that adds the ability to turn off the light by frobbing it.
 * Also supports an emitter tweq that simulates an unignited jet of gas.
 *
 */
DECLARE_SCRIPT(cScr_NewGasLight,cScript)
{
public:
	DECLARE_FACTORY(NewGasLight,HookScript);
	cScr_NewGasLight(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId), m_pExtinguishable(NULL),
		SCRIPT_VAROBJ(cScr_NewGasLight,m_iRealMode,iHostObjId),
		SCRIPT_VAROBJ(cScr_NewGasLight,m_bStartsOn,iHostObjId)
	{ }
	virtual ~cScr_NewGasLight ();

private:
	DECLARE_MSGHANDLER(OnBeginScript);
	DECLARE_MSGHANDLER(OnSim);
	DECLARE_MSGHANDLER(OnTimer);
	DECLARE_MSGHANDLER(OnFrobWorldEnd);
	DECLARE_MSGHANDLER(OnTurnOn);
	DECLARE_MSGHANDLER(OnTurnOff);
	DECLARE_MSGHANDLER(OnToggle);
	DECLARE_MSGHANDLER(OnExtinguish);

	int ToggleEmitter(int iAction);
	void DoTurnOn(void);
	void DoTurnOff(void);
	void DoToggle(void);

	IScript* m_pExtinguishable;
	script_int m_iRealMode;
	script_int m_bStartsOn;

protected:
	virtual void InitScriptData();
	virtual long __stdcall ReceiveMessage(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace);
};


/*************************
 * NewSuspiciousReactions
 *
 * A replacement for ``SuspiciousReactions`` that is more flexible
 * and intelligent.
 *
 */
DECLARE_SCRIPT(cScr_NewSuspicious,cScript)
{
public:
	DECLARE_FACTORY(HighlySuspicious,CustomScript);
	cScr_NewSuspicious(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId), m_bParamsInitted(false),
		SCRIPT_VAROBJ(cScr_NewSuspicious,m_hAlertTimer,iHostObjId),
		SCRIPT_VAROBJ(cScr_NewSuspicious,m_iAlertLevel,iHostObjId),
		SCRIPT_VAROBJ(cScr_NewSuspicious,m_bActionPending,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnBeginScript);
	DECLARE_MSGHANDLER(OnTimer);
	DECLARE_MSGHANDLER(OnSuspicious);
	DECLARE_MSGHANDLER(OnObjActResult);

	void InitParams(void);
	int CountSuspiciousLinks(const char* pszType, int iSuspObj, int *iLikeLinks);
	float CalculateSuspicious(const char* pszType, int iSuspObj, int iTime);
	void HandleSuspicious(const char* pszType, int iSuspObj);
	static int CalcSuspIter(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData);
	void ScanLinks(int iTime, int iSuspObj);
	void ContinueAction(const char* pszAction, int iSuspObj);
	void RaiseAlertLevel(int iLevel, int iTime);
	void RelaxAlertLevel(int iLevel);
	void GotoFrobObject(int iSuspObj);
	void WatchObject(int iSuspObj, int iRelaxTime);
	bool TestIgnoreDoor(int iDoor);

	bool m_bParamsInitted;
	int m_iAwarenessFade;
	float m_fAlertThreshold[3];
	int m_iAlertRelaxTime[3];
	bool m_bGasRelightOnly;

	script_handle<tScrTimer> m_hAlertTimer;
	script_int m_iAlertLevel;
	script_int m_bActionPending;

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("BeginScript",cScr_NewSuspicious::OnBeginScript);
		REGISTER_MSGHANDLER("Timer",cScr_NewSuspicious::OnTimer);
		REGISTER_MSGHANDLER("Suspicious",cScr_NewSuspicious::OnSuspicious);
		REGISTER_MSGHANDLER("ObjActResult",cScr_NewSuspicious::OnObjActResult);
	}
};


/********************
 * NewSecurityDoor
 *
 * Patch for ``SecurityDoor`` that ignores opens by an AI.
 *
 */
DECLARE_SCRIPT(cScr_NewSecurityDoor,cScript)
{
public:
	DECLARE_FACTORY(NewSecurityDoor,HookScript);
	cScr_NewSecurityDoor(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId), m_pSecurityDoor(NULL)
	{ }
	virtual ~cScr_NewSecurityDoor ();

private:
	DECLARE_MSGHANDLER(OnDoorOpenClose);
	IScript* m_pSecurityDoor;

protected:
	virtual void InitScriptData();
	virtual long __stdcall ReceiveMessage(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace);
};


#endif
