/////////////////////////////////////////
// Scripts by Tom N Harris (Telliamed)
// mailto:telliamed@whoopdedo.cjb.net
//

#ifndef TNHSCRIPT_H
#define TNHSCRIPT_H

#if _MSC_VER > 1000
#pragma once
#endif

#include <lg/scrservices.h>
#include <lg/propdefs.h>
#include <darkhook.h>

#include "scriptvars.h"
#include "Script.h"
#include "GenericScript.h"
#include "GenericTrap.h"
#include "GenericScale.h"
#include "GenericControl.h"
#include "GenericTrigger.h"

DECLARE_MODULE(tnhScript)


/********************
 * ScriptController
 *
 * Send a ``ScriptControl`` message to other objects.
 *
 * Inherits: GenericControl
 */
DECLARE_SCRIPT(cScr_ScriptCtrl, cScrControl)
{
public:
	DECLARE_FACTORY(ScriptController,GenericControl);
	cScr_ScriptCtrl(const char* pszName, int iHostObjId)
		: cScrControl(pszName,iHostObjId)
	{ }

  protected:
	virtual long OnControlString(const char*,sScrMsg*,sMultiParm*);
};


/********************
 * CommandControl
 *
 * Execute an arbitrary command.
 *
 * Inherits: GenericControl
 */
DECLARE_SCRIPT(cScr_CommandCtrl, cScrControl)
{
public:
	DECLARE_FACTORY(CommandControl,GenericControl);
	cScr_CommandCtrl(const char* pszName, int iHostObjId)
		: cScrControl(pszName,iHostObjId)
	{ }

  protected:
	virtual long OnControlString(const char*,sScrMsg*,sMultiParm*);
};


/*********************
 * TrapFrobber
 *
 * Translate a ``TurnOn`` message into ``FrobWorldEnd``
 *
 * Inherits: GenericTrap
 * Parameters: inv, player_frob
 */
DECLARE_SCRIPT(cScr_TrapFrobber, cScrTrap)
{
public:
	DECLARE_FACTORY(TrapFrobber, GenericTrap);
	cScr_TrapFrobber(const char* pszName, int iHostObjId)
		: cScrTrap(pszName,iHostObjId)
	{ }

protected:
	virtual long OnTurnOn(sScrMsg* pMsg, sMultiParm* pReply);
};


/*********************
 * TrigFrobber
 *
 * Translate ``FrobWorldEnd`` into an arbitrary message or ``TurnOn``
 *
 * Parameters: inv, frobmsg
 */
DECLARE_SCRIPT(cScr_TrigFrobber, cScript)
{
  public:
	DECLARE_FACTORY(TrigFrobber, CustomScript);
	cScr_TrigFrobber(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId)
	{ }

  private:
	//DECLARE_MSGHANDLER(OnBeginScript);
	DECLARE_MSGHANDLER(OnFrob);

  protected:
	virtual void InitScriptData();
};


/***********************
 * TrapMoveRelative
 *
 * Move and apply velocity to an object relative to its current position
 *
 * Inherits: GenericControl
 * Links: ControlDevice(TrapObject, TargetObject)
 * Parameters: x,y,z, dx,dy,dz, h,p,b, dh,dp,db, rotate_axes
 */
DECLARE_SCRIPT(cScr_MoveRelativeTrap, cScrControl)
{
  public:
	DECLARE_FACTORY(TrapMoveRelative,GenericControl);
	cScr_MoveRelativeTrap(const char* pszName, int iHostObjId)
		: cScrControl(pszName,iHostObjId)
	{ }
  private:
	DECLARE_MSGHANDLER(OnSim);
	static int LinkIterFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData);

  protected:
	virtual long OnControlString(const char*,sScrMsg*,sMultiParm*);
	virtual long OnTurnOn(sScrMsg* pMsg, sMultiParm* pReply);
	virtual long OnTurnOff(sScrMsg* pMsg, sMultiParm* pReply);
	virtual void InitScriptData()
	{
		cScrControl::InitScriptData();
		REGISTER_MSGHANDLER("Sim", cScr_MoveRelativeTrap::OnSim);
	}
};

/**********************
 * TrapCampaignQVar
 * 
 * Set the value for a quest variable that is stored in the campaign database.
 * See documentation of ``TrapSetQVar`` for details.
 * 
 * Inherits: GenericTrap
 * Properties: Trap\QVar
 */
DECLARE_SCRIPT(cScr_CampaignQVarTrap,cScrTrap)
{
  public:
	DECLARE_FACTORY(TrapCampaignQVar,GenericTrap);
	cScr_CampaignQVarTrap(const char* pszName, int iHostObjId)
		: cScrTrap(pszName,iHostObjId)
	{ }

  private:
	DECLARE_MSGHANDLER(OnSim);
	char* GetQVar(char* op, int* val);

  protected:
	virtual long OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply);
	virtual void InitScriptData()
	{
		cScrTrap::InitScriptData();
		REGISTER_MSGHANDLER("Sim", cScr_CampaignQVarTrap::OnSim);
	}
};


/*********************
 * TrapKillEvents
 *
 * Disables user input.  Will automatically enable events after 60 seconds, or 
 * the time specified in [[dromed:property:ScriptTiming|Script\Timing]].
 *
 * The event mask is a binary string representing the events to disable.
 *   * 1000 - Allow mouse movement.
 *   * 0100 - Allow mouse buttons.
 *   * 0010 - ??? (Joystick, maybe)
 *   * 0001 - Allow keyboard.
 *
 * Inherits: GenericTrap
 * Parameters: event_mask
 */
DECLARE_SCRIPT(cScr_KillEvents,cScrTrap)
{
public:
	DECLARE_FACTORY(TrapKillEvents,GenericTrap);
	cScr_KillEvents(const char* pszName, int iHostObjId)
		: cScrTrap(pszName,iHostObjId)
	{ }

private:

protected:
	virtual long OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply);
};


/*************************
 * TrapCamAttach
 *
 * Attach the camera to the script object.
 *
 * Inherits: GenericTrap
 * Parameters: nolens, static, forcereturn
 */
DECLARE_SCRIPT(cScr_CamAttach,cScrTrap)
{
public:
	DECLARE_FACTORY(TrapCamAttach,GenericTrap);
	cScr_CamAttach(const char* pszName, int iHostObjId)
		: cScrTrap(pszName,iHostObjId)
	{ }

protected:
	virtual long OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply);
};


/*********************
 * TrapFreezePlayer
 *
 * Prevent the player from moving.
 *
 * Inherits: GenericScale
 * Parameters: freeze_speed, freeze_name
 */
DECLARE_SCRIPT(cScr_FreezePlayer,cScrScale)
{
public:
	DECLARE_FACTORY(TrapFreezePlayer,GenericScale);
	cScr_FreezePlayer(const char* pszName, int iHostObjId)
		: cScrScale(pszName, iHostObjId),
		  SCRIPT_VAROBJ(cScr_FreezePlayer,m_iIsFrozen,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnBeginScript)
	{
		m_iIsFrozen.Init(0);
		return 1;
	}

	script_int m_iIsFrozen;

protected:
	virtual long OnTurnOn(sScrMsg* pMsg, sMultiParm* pReply);
	virtual long OnTurnOff(sScrMsg* pMsg, sMultiParm* pReply);
	virtual long OnScale(float fValue, sScrMsg* pMsg, sMultiParm* pReply);
	virtual void InitScriptData()
	{
		cScrScale::InitScriptData();
		REGISTER_MSGHANDLER("BeginScript",cScr_FreezePlayer::OnBeginScript);
	}
};


/***********************
 * TrapThreshold
 *
 * Relay ``TurnOn`` every n-th time. ``TurnOff`` decrements the counter unless ''NoOff'' is used.
 * The counter can be global, or tied to the message source.
 * 
 * Inherits: GenericTrap
 * Parameters: threshold, initial, floor, global
 */
DECLARE_SCRIPT(cScr_ThresholdRelay, cScrTrap)
{
  public:
	DECLARE_FACTORY(TrapThreshold,GenericTrap);
	cScr_ThresholdRelay(const char* pszName, int iHostObjId)
		: cScrTrap(pszName,iHostObjId),
		  SCRIPT_VAROBJ(cScr_ThresholdRelay,m_iGlobalCounter,iHostObjId)
	{ }

  private:
	DECLARE_MSGHANDLER(OnBeginScript);
	script_int m_iGlobalCounter;

  protected:
	virtual long OnTurnOn(sScrMsg* pMsg, sMultiParm* pReply);
	virtual long OnTurnOff(sScrMsg* pMsg, sMultiParm* pReply);
	virtual void InitScriptData()
	{
		cScrTrap::InitScriptData();
		REGISTER_MSGHANDLER("BeginScript",cScr_ThresholdRelay::OnBeginScript);
	}
};


/********************
 * TrapRandomTimer
 *
 * Relay ``TurnOn``/``TurnOff`` along ``ControlDevice`` links after a random delay.
 * The time will be at least $$min_time$$ and no more than $$max_time$$.
 * Respects all ``TrapControlFlags``.
 *
 * Parameters: min_time, max_time
 * Messages: TurnOn, TurnOff
 * Links: ControlDevice
 */
DECLARE_SCRIPT(cScr_RandomTimer, cScript)
{
public:
	DECLARE_FACTORY(TrapRandomTimer,CustomScript);
	cScr_RandomTimer(const char* pszName, int iHostObjId)
		: cScript(pszName, iHostObjId),
		  SCRIPT_VAROBJ(cScr_RandomTimer,m_hTimerHandle,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnTurnOnOff);
	DECLARE_MSGHANDLER(OnTimer);
	DECLARE_MSGHANDLER(OnBeginScript)
	{
		m_hTimerHandle.Init();
		return 1;
	}

	script_handle<tScrTimer> m_hTimerHandle;
	void RelayMessage(const char* pszMsgName);

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("BeginScript",cScr_RandomTimer::OnBeginScript);
		REGISTER_MSGHANDLER("TurnOn",cScr_RandomTimer::OnTurnOnOff);
		REGISTER_MSGHANDLER("TurnOff",cScr_RandomTimer::OnTurnOnOff);
		REGISTER_MSGHANDLER("Timer",cScr_RandomTimer::OnTimer);
	}
};


/******************
 * TrapRelayRepeat
 * 
 * Relay ``TurnOn`` a repeated number of times with a delay between firings.
 * ``TurnOff`` will abort unless ''NoOff'' is set. The delay time is milliseconds
 * by default, but can be seconds. If repeat is negative (the default),
 * the relay will continue indefinitely. Starts on its own with autostart.
 * Sorta like a flicker trigger that stops on its own.
 * Respects ``TrapControlFlag`` ''once'' and ''nooff''.
 *
 * Parameters: repeat, interval, autostart
 * Messages: TurnOn, TurnOff
 * Links: ControlDevice
 */
DECLARE_SCRIPT(cScr_RepeatingRelay, cScript)
{
public:
	DECLARE_FACTORY(TrapRelayRepeat,CustomScript);
	cScr_RepeatingRelay(const char* pszName, int iHostObjId)
		: cScript(pszName, iHostObjId), 
		  SCRIPT_VAROBJ(cScr_RepeatingRelay,m_iRepeatCount,iHostObjId),
		  SCRIPT_VAROBJ(cScr_RepeatingRelay,m_hTimerHandle,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnBeginScript)
	{
		m_iRepeatCount.Init();
		m_hTimerHandle.Init();
		return 1;
	}
	DECLARE_MSGHANDLER(OnSim);
	DECLARE_MSGHANDLER(OnTurnOn);
	DECLARE_MSGHANDLER(OnTurnOff);
	DECLARE_MSGHANDLER(OnTimer);

	script_int m_iRepeatCount;
	script_handle<tScrTimer> m_hTimerHandle;

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("BeginScript",cScr_RepeatingRelay::OnBeginScript);
		REGISTER_MSGHANDLER("Sim",cScr_RepeatingRelay::OnSim);
		REGISTER_MSGHANDLER("TurnOn",cScr_RepeatingRelay::OnTurnOn);
		REGISTER_MSGHANDLER("TurnOff",cScr_RepeatingRelay::OnTurnOff);
		REGISTER_MSGHANDLER("Timer",cScr_RepeatingRelay::OnTimer);
	}
};


/*****************
 * TrapRenderFlash
 *
 * Displays a render flash as described by an archetype linked from this object
 * with a ``RenderFlash`` link.
 * Also, if there is a schema linked with ``SoundDescription``, that will be played.
 *
 * One small drawback is that this will cause the current weapon to be deselected.
 *
 * Inherits: GenericTrap
 * Messages: TurnOn
 * Links: RenderFlash(TrapObject, FlashArchetype), SoundDescription(TrapObject, SchemaArchetype)
 */
DECLARE_SCRIPT(cScr_RenderFlash,cScrTrap)
{
  public:
	DECLARE_FACTORY(TrapRenderFlash,GenericTrap);
	cScr_RenderFlash(const char* pszName, int iHostObjId)
		: cScrTrap(pszName,iHostObjId)
	{ }

  private:
	int Relink(ILinkSrv* pLS, ILinkToolsSrv* pLTS, long flavor, int source, int dest);

  protected:
	virtual long OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply);
};


/*******************
 * TrapDisarmPlayer
 *
 * Remove offensive items from the player's inventory.
 * Items are returned to the player when the trap is turned off.
 *
 * Inherits: GenericTrap
 * Parameters: nopotions(1 = don't remove potions)
 */
DECLARE_SCRIPT(cScr_DisarmPlayer,cScrTrap)
{
  public:
	DECLARE_FACTORY(TrapDisarmPlayer,GenericTrap);
	cScr_DisarmPlayer(const char* pszName, int iHostObjId)
		: cScrTrap(pszName,iHostObjId)
	{ }

  private:
	DECLARE_MSGHANDLER(OnContainer);

	static int DisarmIterFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData);

  protected:
	virtual long OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply);

	virtual void InitScriptData()
	{
		cScrTrap::InitScriptData();
		REGISTER_MSGHANDLER("Container",cScr_DisarmPlayer::OnContainer);
	}
};


/*******************
 * ObjFrobInert
 * 
 * Keeps linked objects frob-inerted until the object is ``Slain``.
 * Is also a base script for similar scripts.
 *
 * Messages: Slain, Destroy
 * Links: ScriptParams(Door, FrobableObject, "FrobInert")
 */
DECLARE_SCRIPT(cScr_BaseFrobInert, cScript)
{
public:
	DECLARE_FACTORY(ObjFrobInert,CustomScript);
	cScr_BaseFrobInert(const char* pszName, int iHostObjId)
		: cScript(pszName, iHostObjId),
		  SCRIPT_VAROBJ(cScr_BaseFrobInert,m_iDead,iHostObjId)
	{ }

protected:
	// When the object goes "dead", then the objects are frobbable.
	script_int m_iDead;
	
private:
	DECLARE_MSGHANDLER(OnSim);
	DECLARE_MSGHANDLER(OnSlain);
	DECLARE_MSGHANDLER(OnDestroy);

protected:
	virtual void DoFrobInert(bool bFrobbable);
	virtual bool IsFrobbable(void);

	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("Sim",cScr_BaseFrobInert::OnSim);
		REGISTER_MSGHANDLER("Slain",cScr_BaseFrobInert::OnSlain);
		REGISTER_MSGHANDLER("Destroy",cScr_BaseFrobInert::OnDestroy);
	}

private:
	static int MakeUnfrobbableIterFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData);
	static int MakeFrobbableIterFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData);

};


/*******************
 * DoorFrobInert
 *
 * Add and remove the ``FrobInert`` metaproperty based on the door state of this object.
 *
 * Inherits: ObjFrobInert
 * Messages: DoorOpening, DoorClose, Slain
 * Parameters: breakable - Once the door is slain, it will be treated as always open.
 *             The default is to ignore Slain.
 */
DECLARE_SCRIPT(cScr_DoorFrobInert, cScr_BaseFrobInert)
{
public:
	DECLARE_FACTORY(DoorFrobInert,ObjFrobInert);
	cScr_DoorFrobInert(const char* pszName, int iHostObjId)
		: cScr_BaseFrobInert(pszName, iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnDoorOpening);
	DECLARE_MSGHANDLER(OnDoorClose);
	DECLARE_MSGHANDLER(OnSlain);

protected:
	virtual bool IsFrobbable(void);

	virtual void InitScriptData()
	{
		cScr_BaseFrobInert::InitScriptData();
		REGISTER_MSGHANDLER("DoorOpening",cScr_DoorFrobInert::OnDoorOpening);
		REGISTER_MSGHANDLER("DoorClose",cScr_DoorFrobInert::OnDoorClose);
		REGISTER_MSGHANDLER("Slain",cScr_DoorFrobInert::OnSlain);
	}
};


/*******************
 * GrabFrobInert
 *
 * Removes the ``FrobInert`` metaproperty when the object is picked-up.
 *
 * Inherits: ObjFrobInert
 * Messages: FrobWorldEnd
 */
DECLARE_SCRIPT(cScr_GrabFrobInert, cScr_BaseFrobInert)
{
public:
	DECLARE_FACTORY(GrabFrobInert,ObjFrobInert);
	cScr_GrabFrobInert(const char* pszName, int iHostObjId)
		: cScr_BaseFrobInert(pszName, iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnFrobWorldEnd);

protected:
	virtual void InitScriptData()
	{
		cScr_BaseFrobInert::InitScriptData();
		REGISTER_MSGHANDLER("FrobWorldEnd",cScr_GrabFrobInert::OnFrobWorldEnd);
	}
};


/*******************
 * TweqFrobInert
 *
 * Add and remove the ``FrobInert`` metaproperty based on the state of a joint tweq.
 *
 * Inherits: ObjFrobInert
 * Messages: TweqComplete
 */
DECLARE_SCRIPT(cScr_TweqFrobInert, cScr_BaseFrobInert)
{
public:
	DECLARE_FACTORY(TweqFrobInert,ObjFrobInert);
	cScr_TweqFrobInert(const char* pszName, int iHostObjId)
		: cScr_BaseFrobInert(pszName, iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnTweqComplete);

protected:
	virtual bool IsFrobbable(void);

	virtual void InitScriptData()
	{
		cScr_BaseFrobInert::InitScriptData();
		REGISTER_MSGHANDLER("TweqComplete",cScr_TweqFrobInert::OnTweqComplete);
	}
};


/*******************
 * TrapFrobInert
 *
 * ``TurnOn`` makes the objects frobbable, ``TurnOff`` the opposite.
 * Respects ``TrapControlFlags`` and will do auto-off with ``ScriptTiming``.
 *
 * Inherits: ObjFrobInert
 * Messages: TurnOn, TurnOff, Timer
 */
DECLARE_SCRIPT(cScr_TrapFrobInert, cScr_BaseFrobInert)
{
public:
	DECLARE_FACTORY(TrapFrobInert,ObjFrobInert);
	cScr_TrapFrobInert(const char* pszName, int iHostObjId)
		: cScr_BaseFrobInert(pszName, iHostObjId),
		SCRIPT_VAROBJ(cScr_TrapFrobInert,m_iFrobbable,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnBeginScript)
	{
		m_iFrobbable.Init(0);
		return 1;
	}
	DECLARE_MSGHANDLER(OnTurnOn);
	DECLARE_MSGHANDLER(OnTurnOff);
	DECLARE_MSGHANDLER(OnTimer);

	script_int m_iFrobbable;

protected:
	virtual bool IsFrobbable(void);

	virtual void InitScriptData()
	{
		cScr_BaseFrobInert::InitScriptData();
		REGISTER_MSGHANDLER("BeginScript",cScr_TrapFrobInert::OnBeginScript);
		REGISTER_MSGHANDLER("TurnOn",cScr_TrapFrobInert::OnTurnOn);
		REGISTER_MSGHANDLER("TurnOff",cScr_TrapFrobInert::OnTurnOff);
		REGISTER_MSGHANDLER("Timer",cScr_TrapFrobInert::OnTimer);
	}
};


/***************************
 * UnlinkOnContained
 *
 * When the object is contained, delete all links from another object.
 * The source(s) of the links is determined from a ``ScriptParams`` link with data $$Unlink$$
 * from this object.
 *
 * Messages: Contained
 * Links: ScriptParams(ScriptObject, LinkedObject, "Unlink")
 */
DECLARE_SCRIPT(cScr_ContainedUnlink,cScript)
{
public:
	DECLARE_FACTORY(UnlinkOnContained,CustomScript);
	cScr_ContainedUnlink(const char* pszName, int iHostObjId)
		: cScript(pszName, iHostObjId),
		  SCRIPT_VAROBJ(cScr_ContainedUnlink,m_iAlreadyUnlinked,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnBeginScript)
	{
		m_iAlreadyUnlinked.Init(0);
		return 1;
	}
	DECLARE_MSGHANDLER(OnContained);

	script_int m_iAlreadyUnlinked;

	static int ScriptParamsIterFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData);
	static int DestroyLinksIterFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData);

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("BeginScript",cScr_ContainedUnlink::OnBeginScript);
		REGISTER_MSGHANDLER("Contained",cScr_ContainedUnlink::OnContained);
	}
};


/*******************
 * TimedGrenade
 *
 * Slays the object a specified time after being thrown (not dropped). 
 * Also slays if the object is firestimmed, or if it is poked or bashed while armed.
 * The grenade can be disarmed if you frob it within a safety period,
 * but it will be slain if the timer has less than 1600ms.
 * If a ``RenderFlash`` link is present on the grenade archetype, will trigger a world flash.
 *
 * Will play ``Activate``/``ActiveLoop``/``Deactivate`` schema events.
 * A model tweq will also be activated.
 *
 * Parameters: launch_active
 * Messages: FrobWorldBegin, FrobInvEnd, Slain
 * Stimuli: FireStim, BashStim, PokeStim
 * Properties: Script\Timing, Tweq\Models
 * Links: RenderFlash(GrenadeArchetype, FlashArchetype)
 * Schema Tags: Event Activate, Event Deactivate, Event ActiveLoop, Event Death
 */
DECLARE_SCRIPT(cScr_TimedGrenade,cScript)
{
public:
	DECLARE_FACTORY(TimedGrenade,CustomScript);
	cScr_TimedGrenade(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId),
		  SCRIPT_VAROBJ(cScr_TimedGrenade,m_iSafety,iHostObjId),
		  SCRIPT_VAROBJ(cScr_TimedGrenade,m_hFuseTimer,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnBeginScript);
	DECLARE_MSGHANDLER(OnEndScript);
	DECLARE_MSGHANDLER(OnTimer);
	DECLARE_MSGHANDLER(OnSlain);
	DECLARE_MSGHANDLER(OnFrobInv);
	DECLARE_MSGHANDLER(OnFrobWorld);
	DECLARE_MSGHANDLER(OnContained);
	DECLARE_MSGHANDLER(OnCreate);
	DECLARE_MSGHANDLER(OnStimulus);
	DECLARE_MSGHANDLER(OnSchemaDone);

	script_int m_iSafety;
	script_handle<tScrTimer> m_hFuseTimer;

	void GrenadeFlash(void);
	void GrenadeActive(bool bActivate);
	void GrenadeArmed(void);

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("BeginScript",cScr_TimedGrenade::OnBeginScript);
		REGISTER_MSGHANDLER("EndScript",cScr_TimedGrenade::OnEndScript);
		REGISTER_MSGHANDLER("Timer",cScr_TimedGrenade::OnTimer);
		REGISTER_MSGHANDLER("Slain",cScr_TimedGrenade::OnSlain);
		REGISTER_MSGHANDLER("FrobWorldBegin",cScr_TimedGrenade::OnFrobWorld);
		REGISTER_MSGHANDLER("FrobInvEnd",cScr_TimedGrenade::OnFrobInv);
		REGISTER_MSGHANDLER("Contained",cScr_TimedGrenade::OnContained);
		REGISTER_MSGHANDLER("Create",cScr_TimedGrenade::OnCreate);
		REGISTER_MSGHANDLER("SchemaDone",cScr_TimedGrenade::OnSchemaDone);
		REGISTER_MSGHANDLER("FireStimStimulus",cScr_TimedGrenade::OnStimulus);
		//REGISTER_MSGHANDLER("BashStimStimulus",cScr_TimedGrenade::OnStimulus);
		//REGISTER_MSGHANDLER("PokeStimStimulus",cScr_TimedGrenade::OnStimulus);
	}
};


/********************
 * ZeroGravRoom
 * 
 * Sets the player's gravity to zero when entered. Sets it to 100% on exit.
 * 
 * This is a copy of the script with the same name from System Shock 2.
 * 
 * Messages: PlayerRoomEnter, PlayerRoomExit
 */
DECLARE_SCRIPT(cScr_ZeroGravRoom,cScript)
{
public:
	DECLARE_FACTORY(ZeroGravRoom,CustomScript);
	cScr_ZeroGravRoom(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnEnter);
	DECLARE_MSGHANDLER(OnExit);

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("PlayerRoomEnter",cScr_ZeroGravRoom::OnEnter);
		REGISTER_MSGHANDLER("PlayerRoomExit",cScr_ZeroGravRoom::OnExit);
		//REGISTER_MSGHANDLER("CreatureRoomEnter",cScr_ZeroGravRoom::OnEnter);
		//REGISTER_MSGHANDLER("CreatureRoomExit",cScr_ZeroGravRoom::OnExit);
	}
};


/********************
 * CreatureGravRoom
 * 
 * It doesn't work.
 *
 * Messages: CreatureRoomEnter, CreatureRoomExit
DECLARE_SCRIPT(cScr_CreatGravRoom,cScript)
{
public:
	DECLARE_FACTORY(CreatureGravRoom,CustomScript);
	cScr_CreatGravRoom(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnEnter);
	DECLARE_MSGHANDLER(OnExit);

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("PlayerRoomEnter",cScr_CreatGravRoom::OnEnter);
		REGISTER_MSGHANDLER("PlayerRoomExit",cScr_CreatGravRoom::OnExit);
		REGISTER_MSGHANDLER("CreatureRoomEnter",cScr_CreatGravRoom::OnEnter);
		REGISTER_MSGHANDLER("CreatureRoomExit",cScr_CreatGravRoom::OnExit);
	}
};
 */


/********************
 * TrapFadeOut
 *
 * Fade to black.
 *
 * Inherits: GenericScale
 */
DECLARE_SCRIPT(cScr_FadeOut,cScrScale)
{
public:
	DECLARE_FACTORY(TrapFadeOut,GenericScale);
	cScr_FadeOut(const char* pszName, int iHostObjId)
		: cScrScale(pszName, iHostObjId)
	{ }

protected:
	virtual long OnTurnOff(sScrMsg* pMsg, sMultiParm* pReply);
	virtual long OnScale(float fValue, sScrMsg* pMsg, sMultiParm* pReply);
};


/**************************
 * LockHelper
 *
 * Script that sends ``TurnOn`` based on various interactions with a locked object.
 *
 * Needs [[PickHelper]] on lockpicks and [[KeyHelper]] on keys.
 * 
 * Messages: FrobWorldEnd, UseKey, Wrongkey, PickBegin, PickEnd
 * Links: ScriptParams(data: FrobLock, UseKey, Wrongkey, PickBegin, PickEnd)
 */
DECLARE_SCRIPT(cScr_LockHelper,cScript)
{
public:
	DECLARE_FACTORY(LockHelper,CustomScript);
	cScr_LockHelper(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnFrob);
	DECLARE_MSGHANDLER(OnMessage);
	static int LinkIterFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData);
	static int LinkIterOffFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData);

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("FrobWorldEnd",cScr_LockHelper::OnFrob);
		REGISTER_MSGHANDLER("Wrongkey",cScr_LockHelper::OnMessage);
		REGISTER_MSGHANDLER("UseKey",cScr_LockHelper::OnMessage);
		REGISTER_MSGHANDLER("PickBegin",cScr_LockHelper::OnMessage);
		REGISTER_MSGHANDLER("PickEnd",cScr_LockHelper::OnMessage);
	}
};


DECLARE_SCRIPT(cScr_KeyHelper,cScript)
{
public:
	DECLARE_FACTORY(KeyHelper,CustomScript);
	cScr_KeyHelper(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnFrob);

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("FrobToolEnd",cScr_KeyHelper::OnFrob);
	}
};


DECLARE_SCRIPT(cScr_PickHelper,cScript)
{
public:
	DECLARE_FACTORY(PickHelper,CustomScript);
	cScr_PickHelper(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnFrobInvBegin);
	DECLARE_MSGHANDLER(OnFrobInvEnd);

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("FrobToolBegin",cScr_PickHelper::OnFrobInvBegin);
		REGISTER_MSGHANDLER("FrobToolEnd",cScr_PickHelper::OnFrobInvEnd);
	}
};


/***************************
 * TrapSiblingRelay
 *
 * Facilitates a group of relays that work together. The objects are linked in a
 * ring network using ``ScriptParams`` with data $$Sibling$$. Each object should have 
 * exactly one $$Sibling$$ link from it, and exactly one to it. 
 * Normal operation is that every object will relay a ``TurnOn`` or ``TurnOff`` message 
 * along its ``ControlDevice`` links when any one of the objects receives one. All 
 * trap control flags are respected. With the parameter $$exclusive$$, the object 
 * that receives the ``TurnOn`` or ``TurnOff`` message will tell the other objects to 
 * do the opposite. The parameter may be either "''on''" or "''off''". 
 *
 * To allow all the objects in the network to work together, the object with 
 * the lowest ID is designated as the base object. The trap control flags, 
 * timing, and locked properties of this object are used for the entire network.
 * You should replicate the properties on all the other objects anyway, since
 * IDs have a tendency to drift under certain circumstances.
 *
 * Each object can have a number of ``Owns`` links from it to other objects, such 
 * as levers. When the object receives a message along the network, it will 
 * send an appropriate message to the owned object. Because toggling a lever 
 * causes it to send a message back to this object, the script anticipates 
 * and "swallow" these messages. If the owned objects won't cause extra 
 * messages, you can specify $$no_swallow$$. You cannot have a mix of messages that 
 * will be swallowed and some that won't. If sending a ``TurnOn`` or ``TurnOff`` message 
 * to the owned objects will cause the opposite message to be received by this
 * object, you should set $$invert_levers$$. Again, it's all or nothing. This does 
 * not include the ``Invert`` trap control flag for this object (or rather, the 
 * network the object is in). The script needs to know whether the owned objects 
 * are currently on or off. The default is to assume the levers begin in the 
 * off position. Use $$start_on$$ to assume the levers are already on. Once a 
 * message is received, the state is set to that message.
 *
 * If one of the objects in the network is destroyed, it attempts to self-heal
 * the network. Its sibling links will be connected together. If the object 
 * was the base object, it will transfer the trap control flags, timing, and 
 * locked properties to a new base object. The message ``NewBaseObject`` is used 
 * to inform the rest of the network that the base has changed.
 * 
 * And now, I'm going to bed.
 * 
 * Inherits: GenericTrap
 * Parameters: exclusive(on,off), invert_levers, no_swallow, start_on
 * Links: ScriptParams(Object, Object, "Sibling"), Owns(Relay, Lever), ControlDevice
 * Messages: TurnOn, TurnOff, NowLocked, NowUnlocked, NewBaseObject
 */
DECLARE_SCRIPT(cScr_SiblingRelay,cScrTrap)
{
public:
	DECLARE_FACTORY(TrapSiblingRelay,GenericTrap);
	cScr_SiblingRelay(const char* pszName, int iHostObjId)
		: cScrTrap(pszName, iHostObjId),
		  SCRIPT_VAROBJ(cScr_SiblingRelay,m_iBaseObjectId,iHostObjId),
		  SCRIPT_VAROBJ(cScr_SiblingRelay,m_giSiblingLock,iHostObjId),
		  SCRIPT_VAROBJ(cScr_SiblingRelay,m_iSwallowTurnOn,iHostObjId),
		  SCRIPT_VAROBJ(cScr_SiblingRelay,m_iSwallowTurnOff,iHostObjId),
		  SCRIPT_VAROBJ(cScr_SiblingRelay,m_iCurrentState,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnBeginScript);
	DECLARE_MSGHANDLER(OnDestroy);
	DECLARE_MSGHANDLER(OnNewBaseObject);
	DECLARE_MSGHANDLER(OnNowLocked);
	DECLARE_MSGHANDLER(OnNowUnlocked);

	int FindFirstObject(int iBase);
	void PropogateMessage(const char* pszMsg, int param);
	void ToggleLevers(bool bState);

	static int ToggleIterFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData);

	script_int m_iBaseObjectId;
	// This variable is shared amongst siblings.
	// Its obj id will be set to the base object id, not this object's
	// (unless this is the base object, of course)
	script_int m_giSiblingLock;
	script_int m_iSwallowTurnOn;
	script_int m_iSwallowTurnOff;
	script_int m_iCurrentState;

protected:
	virtual long OnTurnOn(sScrMsg* pMsg, sMultiParm* pReply);
	virtual long OnTurnOff(sScrMsg*,sMultiParm*);
	virtual void InitTrapVars();
	virtual bool IsLocked(sScrMsg* pMsg);
	virtual void SetLock(bool bLock);
	virtual void InitScriptData()
	{
		cScrTrap::InitScriptData();
		REGISTER_MSGHANDLER("BeginScript",cScr_SiblingRelay::OnBeginScript);
		REGISTER_MSGHANDLER("Destroy",cScr_SiblingRelay::OnDestroy);
		REGISTER_MSGHANDLER("NewBaseObject",cScr_SiblingRelay::OnNewBaseObject);
		REGISTER_MSGHANDLER("NowLocked",cScr_SiblingRelay::OnNowLocked);
		REGISTER_MSGHANDLER("NowUnlocked",cScr_SiblingRelay::OnNowUnlocked);
	}
};


/**************************
 * TrigContact
 *
 * Triggers when it comes in contact with another object.
 *
 * Inherits: GenericTrigger
 * Messages: PhysContactCreate, PhysContactDestroy
 * Links: ScriptParams("Contact")
 */
DECLARE_SCRIPT(cScr_ContactTrigger,cScript), cScrTrigger
{
public:
	DECLARE_FACTORY(TrigContact,CustomScript);
	cScr_ContactTrigger(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId), cScrTrigger(iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnBeginScript);
	DECLARE_MSGHANDLER(OnEndScript);
	DECLARE_MSGHANDLER(OnPhysContactCreate);
	DECLARE_MSGHANDLER(OnPhysContactDestroy);
	static int FindContactObject(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData);
	static int FindContactArchetype(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData);
	
protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("BeginScript", cScr_ContactTrigger::OnBeginScript);
		REGISTER_MSGHANDLER("EndScript", cScr_ContactTrigger::OnEndScript);
		REGISTER_MSGHANDLER("PhysContactCreate", cScr_ContactTrigger::OnPhysContactCreate);
		REGISTER_MSGHANDLER("PhysContactDestroy", cScr_ContactTrigger::OnPhysContactDestroy);
	}
};


/**********************
 * TrapText
 *
 * Displays book text.
 *
 * Inherits: GenericTrap
 * Properties: Book\Text, Book\Art
 */
DECLARE_SCRIPT(cScr_TrapText,cScrTrap)
{
  public:
	DECLARE_FACTORY(TrapText,GenericTrap);
	cScr_TrapText(const char* pszName, int iHostObjId)
		: cScrTrap(pszName,iHostObjId)
	{ }

  private:
	DECLARE_MSGHANDLER(OnFrob);

  protected:
	virtual long OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply);
	virtual void InitScriptData()
	{
		cScrTrap::InitScriptData();
		REGISTER_MSGHANDLER("FrobWorldEnd",cScr_TrapText::OnFrob);
		REGISTER_MSGHANDLER("WorldSelect",cScr_TrapText::OnFrob);
	}
};


/*****************************
 * FocusPlaque
 *
 * Display the book text on-screen when the object is hilighted.
 * Will remove the text when unfocused, unless you set ''NoOff''.
 * 
 * Messages: WorldSelect, WorldDeSelect
 * Properties: Book\Text
 */
DECLARE_SCRIPT(cScr_FocusPlaque,cScript)
{
public:
	DECLARE_FACTORY(FocusPlaque,CustomScript);
	cScr_FocusPlaque(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnFocus);
	DECLARE_MSGHANDLER(OnDeFocus);

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("WorldSelect",cScr_FocusPlaque::OnFocus);
		REGISTER_MSGHANDLER("WorldDeSelect",cScr_FocusPlaque::OnDeFocus);
	}
};


/*****************************
 * ContainerPhysics
 *
 * Transfers the velocity of a container to the items it contains when 
 * the container is slain.
 * 
 * Handles death-by-collision stupidly. Reverses the impact vector, 
 * instead of reflecting it. Will scale the magnitude to 60% on a 
 * normal death, or 10% on collision.
 * Not futzing with the velocity has a tendency to throw the objects 
 * into solid space. This seems to prevent that most of the time, 
 * and looks "good enough".
 * 
 * Messages: Slain
 */
DECLARE_SCRIPT(cScr_ContainerPhysics,cScript)
{
public:
	DECLARE_FACTORY(ContainerPhysics,CustomScript);
	cScr_ContainerPhysics(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnSlain);

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("Slain",cScr_ContainerPhysics::OnSlain);
	}
};


/**************************
 * RandomPatrolPath
 *
 * Deletes all but one of the ``AICurrentPatrol`` links.
 * Optionally teleports the object to the destination of the remaining link.
 *
 * Messages: Sim
 * Links: AICurrentPatrol
 * Parameters: teleport_ai
 */
DECLARE_SCRIPT(cScr_RandomPatrolPath,cScript)
{
public:
	DECLARE_FACTORY(RandomPatrolPath,CustomScript);
	cScr_RandomPatrolPath(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnSim);

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("Sim",cScr_RandomPatrolPath::OnSim);
	}
};


/**************************
 * RandomStartingPoint
 *
 * Teleports the starting point to a random location.
 *
 * Messages: Sim
 * Links: ScriptParams w/ Start<diff>
 */
DECLARE_SCRIPT(cScr_RandomStartingPoint,cScript)
{
public:
	DECLARE_FACTORY(RandomStartingPoint,CustomScript);
	cScr_RandomStartingPoint(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnSim);

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("Sim",cScr_RandomStartingPoint::OnSim);
	}
};


/*******************
 * TrigDoorOpening
 *
 * Sends ``GoForward``, ``GoReverse``, and ``TurnOff`` when a door opens, closes, or halts.
 * Respects the ``Invert`` TrapControlFlag.
 *
 * Messages: DoorOpening, DoorClosing, DoorOpen, DoorClose, DoorHalt
 * Links: ControlDevice(door, tweqable)
 */
DECLARE_SCRIPT(cScr_TrigDoorOpening, cScript)
{
public:
	DECLARE_FACTORY(TrigDoorOpening,CustomScript);
	cScr_TrigDoorOpening(const char* pszName, int iHostObjId)
		: cScript(pszName, iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnDoorOpening);
	DECLARE_MSGHANDLER(OnDoorClosing);
	DECLARE_MSGHANDLER(OnDoorHalt);

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("DoorClosing",cScr_TrigDoorOpening::OnDoorClosing);
		REGISTER_MSGHANDLER("DoorOpening",cScr_TrigDoorOpening::OnDoorOpening);
		REGISTER_MSGHANDLER("DoorClose",cScr_TrigDoorOpening::OnDoorHalt);
		REGISTER_MSGHANDLER("DoorOpen",cScr_TrigDoorOpening::OnDoorHalt);
		REGISTER_MSGHANDLER("DoorHalt",cScr_TrigDoorOpening::OnDoorHalt);
	}
};


/**************************
 * InvCarryOver
 *
 * On ``Sim`` termination, creates quest variables representing the objects contained 
 * by this object, or by the ``Player`` object if $$inv_player$$ is set.
 *
 * Messages: Sim
 * Links: ScriptParams w/ CarryOver
 * Parameters: inv_code, inv_player
 */
DECLARE_SCRIPT(cScr_InvCarryOver,cScript)
{
public:
	DECLARE_FACTORY(InvCarryOver,CustomScript);
	cScr_InvCarryOver(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnSim);

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("Sim",cScr_InvCarryOver::OnSim);
	}
};


/**************************
 * InvLoadUp
 *
 * On ``Difficulty`` or ``Sim`` start, iterates over quest variables and creates 
 * objects from them. The objects are added to the contents of the script 
 * object, or to the ``Player`` object if $$inv_player$$ is set.
 *
 * Messages: Sim, Difficulty
 * Parameters: inv_code, inv_player
 * Properties: Difficulty\Script
 */
DECLARE_SCRIPT(cScr_InvLoadUp,cScript)
{
public:
	DECLARE_FACTORY(InvLoadUp,CustomScript);
	cScr_InvLoadUp(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnSim);
	DECLARE_MSGHANDLER(OnDifficulty);

	void DoLoadUp(void);
	void ClearLoadUp(void);

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("Sim",cScr_InvLoadUp::OnSim);
		REGISTER_MSGHANDLER("Difficulty",cScr_InvLoadUp::OnDifficulty);
	}
};


/**************************
 * GenericPotion
 *
 * Works with the standard script ``TimedPotion``.
 * On ``StartPotion``, add a metaproperty to the recipient.
 * Recipient is usually the frobber, or specified by the 
 * optional parameter $$potion_dest$$.
 *
 * Messages: StartPotion
 * Parameters: potion_effect, potion_dest
 */
DECLARE_SCRIPT(cScr_GenericPotion,cScript)
{
public:
	DECLARE_FACTORY(GenericPotion,CustomScript);
	cScr_GenericPotion(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId),
		SCRIPT_VAROBJ(cScr_GenericPotion,m_bActive,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnStartPotion);
	DECLARE_MSGHANDLER(OnDestroy);
	DECLARE_MSGHANDLER(OnBeginScript)
	{
		m_bActive.Init(0);
		return 1;
	}

	script_int m_bActive;

	int GetRecipient(void);

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("BeginScript",cScr_GenericPotion::OnBeginScript);
		REGISTER_MSGHANDLER("StartPotion",cScr_GenericPotion::OnStartPotion);
		REGISTER_MSGHANDLER("Destroy",cScr_GenericPotion::OnDestroy);
	}
};


/**************************
 * InventoryMetaprop
 *
 * While contained by the ``Player``, a meta-property is added to the avatar.
 *
 * Messages: Contained
 * Parameters: inv_effect
 */
DECLARE_SCRIPT(cScr_InvMP,cScript)
{
public:
	DECLARE_FACTORY(InventoryMetaprop,CustomScript);
	cScr_InvMP(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnContained);

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("Contained",cScr_InvMP::OnContained);
	}
};


/*******************
 * TrapContainer
 *
 * Moves its contents to a ``ControlDevice`` linked object.
 * If the parameter $$track_contains$$ is set, then the script will 
 * create an ``Owns`` link to the objects it adds to the container.
 * When turned off, it will only take back the owned objects.
 *
 * Inherits: GenericTrap
 * Links: ControlDevice
 * Parameters: track_contains
 */
DECLARE_SCRIPT(cScr_ContainerTrap,cScrTrap)
{
  public:
	DECLARE_FACTORY(TrapContainer,GenericTrap);
	cScr_ContainerTrap(const char* pszName, int iHostObjId)
		: cScrTrap(pszName,iHostObjId)
	{ }

  private:
	DECLARE_MSGHANDLER(OnSim);
	static int ContainObjects(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData);

  protected:
	virtual long OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply);
	virtual void InitScriptData()
	{
		cScrTrap::InitScriptData();
		REGISTER_MSGHANDLER("Sim",cScr_ContainerTrap::OnSim);
	}
};


/*******************
 * TrapDispenser
 *
 * Creates a new object and places it into a container.
 *
 * Inherits: GenericTrap
 * Parameters: dispense
 * Links: ControlDevice
 */
DECLARE_SCRIPT(cScr_DispenserTrap,cScrTrap)
{
  public:
	DECLARE_FACTORY(TrapDispenser,GenericTrap);
	cScr_DispenserTrap(const char* pszName, int iHostObjId)
		: cScrTrap(pszName,iHostObjId)
	{ }

  private:
	  DECLARE_MSGHANDLER(OnSim);

  protected:
	virtual long OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply);
	virtual void InitScriptData()
	{
		cScrTrap::InitScriptData();
		REGISTER_MSGHANDLER("Sim",cScr_DispenserTrap::OnSim);
	}
};


/**************************
 * TrigPurchase
 *
 * Sends ``TurnOn`` if the object was purchased during loadout.
 *
 * Inherits: GenericTrigger
 */
DECLARE_SCRIPT(cScr_PurchaseTrigger,cScript), cScrTrigger
{
public:
	DECLARE_FACTORY(TrigPurchase,CustomScript);
	cScr_PurchaseTrigger(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId), cScrTrigger(iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnSim);
	DECLARE_MSGHANDLER(OnTimer);
	
protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("Sim", cScr_PurchaseTrigger::OnSim);
		REGISTER_MSGHANDLER("Timer", cScr_PurchaseTrigger::OnTimer);
	}
};


/**************************
 * SimpleGun
 *
 * This is a demo script and may (probably) not be around for long.
 * DO NOT USE THIS SCRIPT
 *
 */
DECLARE_SCRIPT(cScr_Gun,cScript)
{
public:
	DECLARE_FACTORY(SimpleGun,CustomScript);
	cScr_Gun(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnSelect);
	DECLARE_MSGHANDLER(OnDeSelect);
	DECLARE_MSGHANDLER(OnFrobBegin);
	DECLARE_MSGHANDLER(OnFrobEnd);
	
protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("InvSelect", cScr_Gun::OnSelect);
		REGISTER_MSGHANDLER("InvDeSelect", cScr_Gun::OnDeSelect);
		REGISTER_MSGHANDLER("FrobInvBegin", cScr_Gun::OnFrobBegin);
		REGISTER_MSGHANDLER("FrobInvEnd", cScr_Gun::OnFrobEnd);
	}
};


/**************************
 * TurnAtWaypoints
 *
 * Modify and activate a rotate tweq to match the heading of 
 * each waypoint encountered. Respond accordingly to ``Starting``/``Stopping``.
 *
 * Messages: Waypoint, TweqComplete, Starting, Stopping
 */
DECLARE_SCRIPT(cScr_TurnVator,cScript)
{
public:
	DECLARE_FACTORY(TurnAtWaypoints,CustomScript);
	cScr_TurnVator(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId),
		SCRIPT_VAROBJ(cScr_TurnVator,m_vTweqRate,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnBeginScript);
	DECLARE_MSGHANDLER(OnWaypoint);
	DECLARE_MSGHANDLER(OnTweqComplete);
	DECLARE_MSGHANDLER(OnStart);
	DECLARE_MSGHANDLER(OnStop);
	
	script_vec m_vTweqRate;

	bool AdjustTweq(const mxs_vector& currvec, const mxs_vector& destvec);

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("BeginScript", cScr_TurnVator::OnBeginScript);
		REGISTER_MSGHANDLER("MovingTerrainWaypoint", cScr_TurnVator::OnWaypoint);
		REGISTER_MSGHANDLER("TweqComplete", cScr_TurnVator::OnTweqComplete);
		REGISTER_MSGHANDLER("Starting", cScr_TurnVator::OnStart);
		REGISTER_MSGHANDLER("Stopping", cScr_TurnVator::OnStop);
	}
};


/**************************
 * VisibleRelay
 *
 * Pass ``TurnOn``/``TurnOff`` only if a particular object is (or is not)
 * currently rendered.
 * Respects ``TrapControlFlags`` ''Once'', ''Invert''.
 *
 * Messages: TurnOn, TurnOff
 * Links: ScriptParams("Visible")
 */
DECLARE_SCRIPT(cScr_VisibleTrap,cScript)
{
public:
	DECLARE_FACTORY(VisibleRelay,CustomScript);
	cScr_VisibleTrap(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnTrigger);
	DECLARE_MSGHANDLER(OnQuery);

	bool TestVisible(bool bVisible);

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("TurnOn", cScr_VisibleTrap::OnTrigger);
		REGISTER_MSGHANDLER("TurnOff", cScr_VisibleTrap::OnTrigger);
		REGISTER_MSGHANDLER("TestVisible", cScr_VisibleTrap::OnQuery);
	}
};


/**************************
 * ResetTweqState
 *
 * On ``TweqComplete``, the state of the tweq is reset to the values inherited
 * from the archetype. On ``TurnOff``, all tweqs will be reset.
 *
 * Messages: TweqComplete, TurnOff
 */
DECLARE_SCRIPT(cScr_ResetTweqs,cScript)
{
public:
	DECLARE_FACTORY(ResetTweqState,CustomScript);
	cScr_ResetTweqs(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId),
		SCRIPT_VAROBJ(cScr_ResetTweqs,m_vRotInit,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnTweqComplete);
	DECLARE_MSGHANDLER(OnTurnOff);
	DECLARE_MSGHANDLER(OnSim);

	script_vec m_vRotInit;

	void ResetTweq(const char* pszName, int iObj, int iArc, IPropertySrv* pPS);
	void ResetProp(const char* pszName, int iObj, int iArc, IPropertySrv* pPS);
	void ResetAll(int iObj, int iArc, IPropertySrv* pPS);

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("TweqComplete", cScr_ResetTweqs::OnTweqComplete);
		REGISTER_MSGHANDLER("TurnOff", cScr_ResetTweqs::OnTurnOff);
		REGISTER_MSGHANDLER("Sim", cScr_ResetTweqs::OnSim);
	}
};


/*******************
 * RecallDevice
 *
 * When turned on, all CD-linked objects are teleported to
 * the location of the ``Player`` object.
 *
 * This is meant to be used to move a teleport trap so when
 * it is activated later, the Player returns to where he was before.
 *
 * Inherits: GenericTrap
 * Links: ControlDevice
 */
DECLARE_SCRIPT(cScr_RecallDevice,cScrTrap)
{
  public:
	DECLARE_FACTORY(RecallDevice,GenericTrap);
	cScr_RecallDevice(const char* pszName, int iHostObjId)
		: cScrTrap(pszName,iHostObjId)
	{ }

  private:
	static int RecallIter(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData);

  protected:
	virtual long OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply);
};


/**************************
 * TrigProxFrob
 *
 * When frobbed, ``TurnOn`` will be sent if there is a ``Population`` link
 * from a specific object (determined via a ``Route`` link) to the frobber.
 *
 * ``Population`` links are created by bounds and room triggers. So you can
 * use this script to relay based on whether or not the Player is standing in 
 * a particular place.
 *
 * Inherits: GenericTrigger
 * Messages: FrobWorldEnd, FrobInvEnd
 * Links: Route
 */
DECLARE_SCRIPT(cScr_ProxFrobTrigger,cScript), cScrTrigger
{
public:
	DECLARE_FACTORY(TrigProxFrob,CustomScript);
	cScr_ProxFrobTrigger(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId), cScrTrigger(iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnFrob);
	
protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("FrobWorldEnd", cScr_ProxFrobTrigger::OnFrob);
		REGISTER_MSGHANDLER("FrobInvEnd", cScr_ProxFrobTrigger::OnFrob);
	}
};


/*******************
 * PoliteVOTrap
 *
 * Plays a VO schema only if there are no ``AIInvest`` links to the
 * ``Player`` object.
 *
 * Inherits: GenericTrap
 * Links: SoundDescription
 */
DECLARE_SCRIPT(cScr_VOPolite,cScrTrap)
{
  public:
	DECLARE_FACTORY(PoliteVOTrap,GenericTrap);
	cScr_VOPolite(const char* pszName, int iHostObjId)
		: cScrTrap(pszName,iHostObjId)
	{ }

  protected:
	virtual long OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply);
};


/**************************
 * TrigSchemaDoneSpec
 *
 * Triggers on ``SchemaDone`` with a specific schema name
 *
 * Inherits: GenericTrigger
 * Messages: SchemaDone
 * Parameters: schema
 */
DECLARE_SCRIPT(cScr_SpecSchemaTrigger,cScript), cScrTrigger
{
public:
	DECLARE_FACTORY(TrigSchemaDoneSpec,CustomScript);
	cScr_SpecSchemaTrigger(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId), cScrTrigger(iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnSchemaDone);
	
protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("SchemaDone", cScr_SpecSchemaTrigger::OnSchemaDone);
	}
};


/**************************
 * AIGoMissing
 *
 * When slain or knocked-out, a marker is created at the
 * idling origin of the AI. The archetype ``MissingPerson`` is used 
 * and should be created with an appropriate suspicious property.
 * The marker is also created when the AI goes into high-alert.
 * and an ``AIWatchObj`` link is created from this AI to the marker object.
 * The marker should have watch link defaults that, when the AI returns
 * to the marker with a lower alert level, then it will send the script
 * message ``AIRelax``. Then the object will be destroyed
 *
 * Messages: AIModeChange, AIAlertness, AIRelax
 * Parameters: missingonalert, missingpatrol
 */
DECLARE_SCRIPT(cScr_MissingPerson,cScript)
{
public:
	DECLARE_FACTORY(AIGoMissing,CustomScript);
	cScr_MissingPerson(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId),
		SCRIPT_VAROBJ(cScr_MissingPerson,m_iMarkerObj,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnBeginScript)
	{
		m_iMarkerObj.Init(0);
		return 1;
	}
	DECLARE_MSGHANDLER(OnTimer);
	DECLARE_MSGHANDLER(OnModeChange);
	DECLARE_MSGHANDLER(OnAlertness);
	DECLARE_MSGHANDLER(OnRelax);

	script_int m_iMarkerObj;

	int CreateMissingMarker(void);
	void PatrolMissing(int iTrolPt);
	int GoMissing(void);
	void NotMissing(int iMarker);
	int DistanceToTime(int iFrom, int iTo);

	static const sAIWatchPoint sm_AIWatchLinkDefault;
	
protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("BeginScript", cScr_MissingPerson::OnBeginScript);
		REGISTER_MSGHANDLER("Timer", cScr_MissingPerson::OnTimer);
		REGISTER_MSGHANDLER("AIModeChange", cScr_MissingPerson::OnModeChange);
		REGISTER_MSGHANDLER("Alertness", cScr_MissingPerson::OnAlertness);
		REGISTER_MSGHANDLER("AIRelax", cScr_MissingPerson::OnRelax);
	}
};


/**************************
 * MissingMarker
 *
 * Delayed response when an object goes missing. 
 * Makes the object suspicious after a timer.
 *
 * Messages: GoMissing, NotMissing
 * Parameters: missing_delay(time)
 */
DECLARE_SCRIPT(cScr_MissingMarker,cScript)
{
public:
	DECLARE_FACTORY(MissingMarker,CustomScript);
	cScr_MissingMarker(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId),
		SCRIPT_VAROBJ(cScr_MissingMarker,m_hDelay,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnBeginScript)
	{
		m_hDelay.Init();
		return 1;
	}
	DECLARE_MSGHANDLER(OnTimer);
	DECLARE_MSGHANDLER(OnGoMissing);
	DECLARE_MSGHANDLER(OnNotMissing);

	script_handle<tScrTimer> m_hDelay;

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("BeginScript", cScr_MissingMarker::OnBeginScript);
		REGISTER_MSGHANDLER("Timer", cScr_MissingMarker::OnTimer);
		REGISTER_MSGHANDLER("GoMissing", cScr_MissingMarker::OnGoMissing);
		REGISTER_MSGHANDLER("NotMissing", cScr_MissingMarker::OnNotMissing);
	}
};


/**************************
 * UndeadRadar
 *
 * React to ``UndeadStim`` stimuli by adjusting the ``SelfLit`` and ``LightColor`` 
 * properties proportionally to the intensity of the stim.
 * The ``Script\Timing`` property controls how long a stim will be effective,
 * during which any lower stims will be squelched.
 *
 * Stimuli: UndeadStim
 * Properties: Renderer\Dynamic Light, Script\Timing
 * Parameters: 
 */
DECLARE_SCRIPT(cScr_UndeadRadar,cScript)
{
public:
	DECLARE_FACTORY(UndeadRadar,CustomScript);
	cScr_UndeadRadar(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId),
		SCRIPT_VAROBJ(cScr_UndeadRadar,m_iIntensity,iHostObjId),
		SCRIPT_VAROBJ(cScr_UndeadRadar,m_iStimTime,iHostObjId),
		SCRIPT_VAROBJ(cScr_UndeadRadar,m_fLastStim,iHostObjId),
		SCRIPT_VAROBJ(cScr_UndeadRadar,m_hRadarTimer,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnBeginScript);
	DECLARE_MSGHANDLER(OnEndScript);
	DECLARE_MSGHANDLER(OnTimer);
	DECLARE_MSGHANDLER(OnSim);
	DECLARE_MSGHANDLER(OnStimulus);

	// Default intensity
	script_int m_iIntensity;
	// Time of last change
	script_int m_iStimTime;
	// Level of last received stim (even if its ignored)
	script_float m_fLastStim;
	// Currently active scale. (Initialized from m_fLastStim)
	float m_fCurrent;
	// Slow fade timer
	script_handle<tScrTimer> m_hRadarTimer;

	bool SetLight(float fScale, int iTime, bool bPersist = false);

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("BeginScript", cScr_UndeadRadar::OnBeginScript);
		REGISTER_MSGHANDLER("EndScript", cScr_UndeadRadar::OnEndScript);
		REGISTER_MSGHANDLER("Timer", cScr_UndeadRadar::OnTimer);
		REGISTER_MSGHANDLER("Sim", cScr_UndeadRadar::OnSim);
		REGISTER_MSGHANDLER("UndeadStimStimulus", cScr_UndeadRadar::OnStimulus);
	}
};


/**************************
 * TrigBreadcrumb
 *
 * Creates regions in which a passing AI can "sense" the Player.
 * Place these in hallway junctions with adjacent traps linked.
 * The Player's movement through the trap is relayed to patrolling AI.
 *
 * Messages: PhysEnter, PhysExit, Breadcrumb(data1 = dest_obj)
 * Properties: Script\Timing
 * Links: Route
 */
DECLARE_SCRIPT(cScr_TrigBreadcrumb,cScript)
{
public:
	DECLARE_FACTORY(TrigBreadcrumb,CustomScript);
	cScr_TrigBreadcrumb(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId),
		SCRIPT_VAROBJ(cScr_TrigBreadcrumb,m_iCurrentRoute,iHostObjId),
		SCRIPT_VAROBJ(cScr_TrigBreadcrumb,m_hTimer,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnBeginScript);
	DECLARE_MSGHANDLER(OnEndScript);
	DECLARE_MSGHANDLER(OnPhysEnter);
	DECLARE_MSGHANDLER(OnPhysExit);
	DECLARE_MSGHANDLER(OnTimer);

	script_int m_iCurrentRoute;
	script_handle<tScrTimer> m_hTimer;

	int DiscernRoute(int iPlayer);

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("BeginScript", cScr_TrigBreadcrumb::OnBeginScript);
		REGISTER_MSGHANDLER("EndScript", cScr_TrigBreadcrumb::OnEndScript);
		REGISTER_MSGHANDLER("PhysEnter", cScr_TrigBreadcrumb::OnPhysEnter);
		REGISTER_MSGHANDLER("PhysExit", cScr_TrigBreadcrumb::OnPhysExit);
		REGISTER_MSGHANDLER("Timer", cScr_TrigBreadcrumb::OnTimer);
	}
};


/**************************
 * FollowBreadcrumb
 *
 * Responds to the message ``Breadcrumb``. The AI will be directed to 
 * go to the object the message specifies. If the AI is at high alert,
 * then nothing happens. The minimum alert level will be set to 1 for
 * 5 seconds.
 *
 * Messages: Breadcrumb(data1 = dest_obj)
 */
DECLARE_SCRIPT(cScr_FollowBreadcrumb,cScript)
{
public:
	DECLARE_FACTORY(FollowBreadcrumb,CustomScript);
	cScr_FollowBreadcrumb(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnTimer);
	DECLARE_MSGHANDLER(OnNotify);

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("Timer", cScr_FollowBreadcrumb::OnTimer);
		REGISTER_MSGHANDLER("Breadcrumb", cScr_FollowBreadcrumb::OnNotify);
	}
};


/**************************
 * RainSFX
 *
 * Turn off particles when weather effects are disabled.
 *
 * Messages: Sim, DarkGameModeChange
 */
DECLARE_SCRIPT(cScr_RainSFX,cScript)
{
public:
	DECLARE_FACTORY(RainSFX,CustomScript);
	cScr_RainSFX(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnSim);

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("Sim", cScr_RainSFX::OnSim);
		REGISTER_MSGHANDLER("DarkGameModeChange", cScr_RainSFX::OnSim);
	}
};


/**************************
 * TrackAttach
 *
 * Adjusts the position of an attached object to follow 
 * the moving base object.
 *
 * Messages: DHNotify
 * Links: PhysAttach
 * Properties: Physics\Model\State
 */
DECLARE_SCRIPT(cScr_TrackAttach, cScript)
{
public:
	DECLARE_FACTORY(TrackAttach,CustomScript);
	cScr_TrackAttach(const char* pszName, int iHostObjId)
		: cScript(pszName, iHostObjId),
		  SCRIPT_VAROBJ(cScr_TrackAttach,m_bInitted,iHostObjId)
	{
		DarkHookInitializeService(g_pScriptManager, g_pMalloc);
	}

private:
	script_int m_bInitted;

	DECLARE_MSGHANDLER(OnBeginScript);
	DECLARE_MSGHANDLER(OnEndScript);
	DECLARE_MSGHANDLER(OnNotify);

	void SetupTrackingLinks(void);
	static int LinkIter(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData);

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("BeginScript",cScr_TrackAttach::OnBeginScript);
		REGISTER_MSGHANDLER("EndScript",cScr_TrackAttach::OnEndScript);
		REGISTER_MSGHANDLER("DHNotify",cScr_TrackAttach::OnNotify);
	}
};


/**************************
 * ElevatorCar
 *
 * When the lift is about to move, activate the joint tweq in reverse and wait for 
 * it to complete. On ``TweqComplete``, activate the lift. On ``WaypointReached``, stop the 
 * car and activate the joint tweq forward. Posts ``TurnOff`` to links from the waypoint 
 * it's leaving and ``TurnOn`` to links from the waypoint it arrives at. On ``TurnOn``, moves
 * to the next waypoint; ``TurnOff`` moves to the previous waypoint.
 *
 * Messages: TurnOn, TurnOff, TweqComplete, MovingTerrainWaypoint, Call, Starting, Stopping
 * Properties: MovingTerrain, Tweq\Joints
 * Links: TPath, TPathNext, ControlDevice, ScriptParams
 */
DECLARE_SCRIPT(cScr_ElevatorCar,cScript)
{
public:
	DECLARE_FACTORY(ElevatorCar,CustomScript);
	cScr_ElevatorCar(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId),
		SCRIPT_VAROBJ(cScr_ElevatorCar,m_iPathDest,iHostObjId),
		SCRIPT_VAROBJ(cScr_ElevatorCar,m_iPathDir,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnBeginScript)
	{
		m_iPathDest.Init();
		m_iPathDir.Init();
		return 1;
	}
	DECLARE_MSGHANDLER(OnSim);
	DECLARE_MSGHANDLER(OnTimer);
	DECLARE_MSGHANDLER(OnTweqComplete);
	DECLARE_MSGHANDLER(OnWaypoint);
	DECLARE_MSGHANDLER(OnStarting);
	DECLARE_MSGHANDLER(OnStopping);
	DECLARE_MSGHANDLER(OnTurnOn);
	DECLARE_MSGHANDLER(OnTurnOff);
	DECLARE_MSGHANDLER(OnCall);

	script_int m_iPathDest;
	script_int m_iPathDir;

	bool TweqControl(int iTweqFlags);
	int CurrentWaypoint(void);
	int FindPath(int iDest);
	bool CanGoForward(void);
	bool CanGoBackward(void);
	bool OpenDoors(void);
	bool CloseDoors(void);
	bool GoForward(void);
	bool GoBackward(void);
	void Halt(void);
	bool IsMoving(void);

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("BeginScript", cScr_ElevatorCar::OnBeginScript);
		REGISTER_MSGHANDLER("Sim", cScr_ElevatorCar::OnSim);
		REGISTER_MSGHANDLER("Timer", cScr_ElevatorCar::OnTimer);
		REGISTER_MSGHANDLER("TweqComplete", cScr_ElevatorCar::OnTweqComplete);
		REGISTER_MSGHANDLER("MovingTerrainWaypoint", cScr_ElevatorCar::OnWaypoint);
		REGISTER_MSGHANDLER("Starting", cScr_ElevatorCar::OnStarting);
		REGISTER_MSGHANDLER("Stopping", cScr_ElevatorCar::OnStopping);
		REGISTER_MSGHANDLER("TurnOn", cScr_ElevatorCar::OnTurnOn);
		REGISTER_MSGHANDLER("TurnOff", cScr_ElevatorCar::OnTurnOff);
		REGISTER_MSGHANDLER("Call", cScr_ElevatorCar::OnCall);
	}
};


/**************************
 * TrigElevator
 *
 * Triggers when moving terrain starts and stops.
 *
 * Messages: Starting, Stopping
 */
DECLARE_SCRIPT(cScr_ElevatorTrigger,cScript)
{
public:
	DECLARE_FACTORY(TrigElevator,CustomScript);
	cScr_ElevatorTrigger(const char* pszName, int iHostObjId)
		: cScript(pszName,iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnStarting);
	DECLARE_MSGHANDLER(OnStopping);
	
protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("Starting", cScr_ElevatorTrigger::OnStarting);
		REGISTER_MSGHANDLER("Stopping", cScr_ElevatorTrigger::OnStopping);
	}
};


/*******************
 * TrapPhysical
 *
 * Control an object's presence in the world.
 *
 * Inherits: GenericTrap
 * Properties: Renderer\HasRefs
 */
DECLARE_SCRIPT(cScr_SetRefs,cScrTrap)
{
  public:
	DECLARE_FACTORY(TrapPhysical,GenericTrap);
	cScr_SetRefs(const char* pszName, int iHostObjId)
		: cScrTrap(pszName,iHostObjId)
	{ }

  protected:
	virtual long OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply);
};


#if 0
/*******************
 * TrigActiveFlicker
 *
 * Trigger for activating devices as long as a flicker tweq is running.
 * Use with $$SimLargeRadius$$ or $$SimSmallRadius$$ to turn things off when
 * the ``Player`` isn't nearby.
 *
 * An initial ``TurnOn`` or ``TurnOff`` message is sent when the game starts.
 * Additionally, the script will respond to ``TurnOn`` and ``TurnOff`` itself to
 * start and stop the tweq manually.
 *
 * Inherits: GenericTrigger
 * Messages: TurnOn, TurnOff, TweqComplete
 * Properties: Tweq\Flicker
 */
DECLARE_SCRIPT(cScr_TweqActive,cScrTrap), cScrTrigger
{
  public:
	DECLARE_FACTORY(TrigActiveFlicker,GenericTrap);
	cScr_TweqActive(const char* pszName, int iHostObjId)
		: cScrTrap(pszName,iHostObjId), cScrTrigger(iHostObjId),
		  SCRIPT_VAROBJ(cScr_TweqActive,m_iActive,iHostObjId)
	{ }

  private:
	DECLARE_MSGHANDLER(OnSim);
	DECLARE_MSGHANDLER(OnTimer);
	DECLARE_MSGHANDLER(OnTweqComplete);
	DECLARE_MSGHANDLER(OnBeginScript)
	{
		m_iActive.Init(-1);
		return 1;
	}
	script_int m_iActive;

  protected:
	virtual long OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply);
	virtual void InitScriptData()
	{
		cScrTrap::InitScriptData();
		REGISTER_MSGHANDLER("BeginScript", cScr_TweqActive::OnBeginScript);
		REGISTER_MSGHANDLER("Sim", cScr_TweqActive::OnSim);
		REGISTER_MSGHANDLER("Timer", cScr_TweqActive::OnTimer);
		REGISTER_MSGHANDLER("TweqComplete", cScr_TweqActive::OnTweqComplete);
	}
};
#endif


/*******************
 * TrigProximity
 *
 * Triggers when the ``Player`` is close to the object.
 *
 * An initial ``TurnOn`` or ``TurnOff`` message is sent when the game starts.
 * Additionally, the script will respond to ``TurnOn`` and ``TurnOff`` itself to
 * start and stop the tweq manually.
 *
 * Inherits: GenericTrigger
 * Messages: TurnOn, TurnOff
 * Parameters: proximity
 */
DECLARE_SCRIPT(cScr_ProxTrig,cScrTrap), cScrTrigger
{
  public:
	DECLARE_FACTORY(TrigProximity,GenericTrap);
	cScr_ProxTrig(const char* pszName, int iHostObjId)
		: cScrTrap(pszName,iHostObjId), cScrTrigger(iHostObjId),
		  SCRIPT_VAROBJ(cScr_ProxTrig,m_fLastDist,iHostObjId),
		  SCRIPT_VAROBJ(cScr_ProxTrig,m_hTimer,iHostObjId)
	{ }

  private:
	DECLARE_MSGHANDLER(OnSim);
	DECLARE_MSGHANDLER(OnTimer);
	DECLARE_MSGHANDLER(OnBeginScript)
	{
		m_fLastDist.Init(-1);
		m_hTimer.Init();
		return 1;
	}
	script_float m_fLastDist;
	script_handle<tScrTimer> m_hTimer;

  protected:
	virtual long OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply);
	virtual void InitScriptData()
	{
		cScrTrap::InitScriptData();
		REGISTER_MSGHANDLER("BeginScript", cScr_ProxTrig::OnBeginScript);
		REGISTER_MSGHANDLER("Sim", cScr_ProxTrig::OnSim);
		REGISTER_MSGHANDLER("Timer", cScr_ProxTrig::OnTimer);
	}
};


/*********************
 * ScaleJointTweq
 *
 * Inherits: GenericScale
 * Properties: Tweq\Joints, Shape\Joint Positions
 * Parameters: scalejoint
 */
DECLARE_SCRIPT(cScr_JointControl,cScrScale)
{
public:
	DECLARE_FACTORY(ScaleJointTweq,GenericScale);
	cScr_JointControl(const char* pszName, int iHostObjId)
		: cScrScale(pszName, iHostObjId)
	{ }

private:
	DECLARE_MSGHANDLER(OnBeginScript)
	{
		return 1;
	}

protected:
	virtual long OnScale(float fValue, sScrMsg* pMsg, sMultiParm* pReply);
	virtual void InitScriptData()
	{
		cScrScale::InitScriptData();
		REGISTER_MSGHANDLER("BeginScript",cScr_JointControl::OnBeginScript);
	}
};


#endif
